import numpy as np
from numpy import ndarray
import sympy as sy

from .utils import Gram

__all__ = [
    "has_full_row_rank",
    "has_full_column_rank",
    "has_full_rank",
    "is_rectangular_frame",
    "is_normal_frame",
    "is_orthonormal_frame",
    "is_independent_frame",
    "is_hermitian",
    "is_pos_def",
    "is_pos_semidef",
]


def is_pos_def(arr) -> bool:
    """
    Returns True if the input is positive definite.
    """
    return np.all(np.linalg.eigvals(arr) > 0)


def is_pos_semidef(arr) -> bool:
    """
    Returns True if the input is positive semi definite.
    """
    return np.all(np.linalg.eigvals(arr) >= 0)


def is_rectangular_frame(axes: ndarray) -> bool:
    """
    Returns True if a frame is Cartesian.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    assert len(axes.shape) == 2, "Input is not a matrix!"
    assert axes.shape[0] == axes.shape[1], "Input is not a square matrix!"
    agram = np.abs(axes @ axes.T)
    return np.isclose(np.trace(agram), np.sum(agram))


def is_normal_frame(axes: ndarray) -> bool:
    """
    Returns True if a frame is normal, meaning, that it's base vectors
    are all of unit length.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return np.allclose(np.linalg.norm(axes, axis=1), 1.0)


def is_orthonormal_frame(axes: ndarray) -> bool:
    """
    Returns True if a frame is orthonormal.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return is_rectangular_frame(axes) and is_normal_frame(axes)


def is_independent_frame(axes: ndarray, tol: float = 0) -> bool:
    """
    Returns True if a the base vectors of a frame are linearly independent.

    Parameters
    ----------
    axes: numpy.ndarray
        A matrix where the i-th row is the i-th basis vector.
    """
    return np.linalg.det(Gram(axes)) > tol


def is_hermitian(arr: ndarray) -> bool:
    """
    Returns True if the input is a hermitian array.
    """
    shp = arr.shape
    s0 = shp[0]
    return all([s == s0 for s in shp[1:]])


def has_full_row_rank(matrix: ndarray) -> bool:
    """
    Returns `True` if the input matrix has full row rank, ie
    if all its rows are linearly independent.

    See also
    --------
    :func:`numpy.linalg.matrix_rank`
    """
    num_rows, num_columns = matrix.shape
    if num_rows > num_columns:
        return False
    rank = np.linalg.matrix_rank(matrix)
    return rank == num_rows


def has_full_column_rank(matrix: ndarray) -> bool:
    """
    Returns `True` if the input matrix has full column rank, ie
    if all its columns are linearly independent.

    See also
    --------
    :func:`numpy.linalg.matrix_rank`
    """
    num_rows, num_columns = matrix.shape
    if num_columns > num_rows:
        return False
    rank = np.linalg.matrix_rank(matrix)
    return rank == num_columns


def has_full_rank(matrix: ndarray | sy.Matrix) -> bool:
    """
    Returns `True` if the input matrix has full rank, `False` otherwise.

    Parameters
    ----------
    matrix : numpy.ndarray | sympy.Matrix
        The input matrix.
    """
    if isinstance(matrix, sy.Matrix):
        return matrix.rank() == min(matrix.shape)
    else:
        return np.linalg.matrix_rank(matrix) == min(matrix.shape)



from typing import Callable, Any
import numpy as np


__all__ = ["squeeze"]


def _squeeze_if_array(arr: Any) -> Any:
    return np.squeeze(arr) if isinstance(arr, np.ndarray) else arr


def squeeze(default: bool = True) -> Callable:
    """
    A decorator that squeezes outputs of a function if
    * the result is a NumPy array
    * the result is a tuple of NumPy arrays
    * the result is a dictionary of NumPy arrays as values

    ```python
    from sigmaepsilon.math import squeeze
    import numpy as np

    @squeeze(default=True)
    def foo(arr):
        return arr

    foo(np.array([[1, 2]]))  # array([1, 2])
    foo(np.array([[1, 2]]), squeeze=False)  # array([[1, 2]])
    ```
    """

    def decorator(fnc: Callable):
        def inner(*args, squeeze: bool = default, **kwargs):
            if squeeze:
                res = fnc(*args, **kwargs)
                if isinstance(res, tuple):
                    return list(map(_squeeze_if_array, res))
                elif isinstance(res, dict):
                    return {k: _squeeze_if_array(v) for k, v in res.items()}
                else:
                    return _squeeze_if_array(res)
            else:
                return fnc(*args, **kwargs)

        inner.__doc__ = fnc.__doc__
        return inner

    return decorator



from sympy import Expr

from sigmaepsilon.core.abstract import ABCMeta_Weak
from .symutils import symbolize


class ABCMeta_MetaFunction(ABCMeta_Weak):
    """
    Metaclass for defining ABCs for algebraic structures.
    """

    def __new__(metaclass, name, bases, namespace, /, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, **kwargs)
        if "value" in namespace:
            cls.f = namespace["value"]

        if "gradient" in namespace:
            cls.g = namespace["gradient"]

        if "Hessian" in namespace:
            cls.G = namespace["Hessian"]

        return cls


class MetaFunction(metaclass=ABCMeta_MetaFunction):
    __slots__ = ("f0", "f1", "f2", "dimension", "domain", "expr", "variables", "vmap")

    def __call__(self, *args, **kwargs):
        return self.f(*args, **kwargs)

    def f(self, *args, **kwargs):
        """
        Returns the function value.

        For this operation the object must have an implementation of

        value(self, *args, **kwargs):
            <...>
            return <...>
        """
        return self.f0(*args, **kwargs)

    def g(self, *args, **kwargs):
        """
        Returns the gradient vector if available.

        For this operation the object must have an implementation of

        gradient(self, *args, **kwargs):
            <...>
            return <...>
        """
        return self.f1(*args, **kwargs)

    def G(self, *args, **kwargs):
        """
        Returns the Hessian matrix if available.

        For this operation the object must have an implementation of

        Hessian(self,*args,**kwargs):
            <...>
            return <...>
        """
        return self.f2(*args, **kwargs)

    @classmethod
    def _str_to_func(cls, str_expr: str, *args, **kwargs):
        return symbolize(*args, str_expr=str_expr, **kwargs)

    @classmethod
    def _sympy_to_func(cls, expr: Expr, *args, **kwargs):
        return symbolize(*args, expr=expr, **kwargs)
